from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, MLClass
from arkindex.ponos.models import Farm
from arkindex.process.models import Process, ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestRetrieveElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.private_corpus = Corpus.objects.create(name="private", public=False)
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = WorkerRun.objects.filter(version=cls.worker_version, process__mode=ProcessMode.Workers).get()

    def setUp(self):
        super().setUp()
        self.page = self.corpus.elements.get(name="Volume 1, page 1r")

    def test_get_element(self):
        ml_class = MLClass.objects.create(name="text", corpus=self.corpus)
        classification = self.vol.classifications.create(worker_version=self.worker_version, ml_class=ml_class)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.vol.id),
            "name": "Volume 1",
            "type": "volume",
            "corpus": {
                "id": str(self.corpus.id),
                "name": "Unit Tests",
                "public": True,
            },
            "thumbnail_url": self.vol.thumbnail.s3_url,
            "thumbnail_put_url": None,
            "worker_version": None,
            "confidence": None,
            "zone": None,
            "rotation_angle": 0,
            "mirrored": False,
            "created": "2020-02-02T01:23:45.678000Z",
            "creator": None,
            "rights": ["read"],
            "metadata_count": 0,
            "classifications": [
                {
                    "id": str(classification.id),
                    "confidence": None,
                    "high_confidence": False,
                    "state": "pending",
                    "worker_version": str(self.worker_version.id),
                    "worker_run": None,
                    "ml_class": {
                        "id": str(ml_class.id),
                        "name": "text",
                    }
                }
            ],
            "worker_run": None
        })

    def test_get_element_zone(self):
        self.assertIsNotNone(self.page.image)
        self.assertIsNotNone(self.page.polygon)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.page.id)}))

        self.assertDictEqual(response.json()["zone"], {
            "id": str(self.page.id),
            "image": {
                "id": str(self.page.image.id),
                "path": "img1",
                "url": "http://server/img1",
                "s3_url": None,
                "width": 1000,
                "height": 1000,
                "status": "unchecked",
                "server": {
                    "display_name": "Test Server",
                    "max_width": None,
                    "max_height": None,
                    "url": "http://server"
                }
            },
            "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
            "url": "http://server/img1/0,0,1000,1000/full/0/default.jpg"
        })

    def test_get_element_thumbnail_folder(self):
        """
        Check getting an element only gives a thumbnail URL with folders
        """
        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["thumbnail_url"], self.vol.thumbnail.s3_url)
        self.assertIsNone(response.json()["thumbnail_put_url"])

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()["thumbnail_url"])
        self.assertIsNone(response.json()["thumbnail_put_url"])

    @override_settings(ARKINDEX_TASKS_IMAGE="task_image")
    def test_get_element_thumbnail_put_ponos_task(self):
        """
        Check getting an element only gives a thumbnail PUT URL Ponos tasks
        running the thumbnails generation command on a folder element
        """
        process = Process.objects.create(
            mode=ProcessMode.Repository,
            revision=self.worker_version.revision,
            creator=self.user,
            generate_thumbnails=True,
            farm=Farm.objects.first(),
        )
        process.run()
        task = process.tasks.get()
        task.image = "task_image"
        task.command = "python generate_thumbnails"

        task.save()
        response = self.client.get(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            HTTP_AUTHORIZATION=f"Ponos {task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["thumbnail_put_url"], self.vol.thumbnail.s3_put_url)

        self.assertFalse(self.page.type.folder)
        response = self.client.get(
            reverse("api:element-retrieve", kwargs={"pk": str(self.page.id)}),
            HTTP_AUTHORIZATION=f"Ponos {task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()["thumbnail_put_url"])

    def test_get_element_thumbnail_put_requires_thumbnails_task(self):
        """
        Only tasks that are intended to generate thumbnails (ARKINDEX_TASKS_IMAGE + thumbnails_generation command)
        can retrieve the thumbnails PUT URL.
        """
        process = Process.objects.create(
            mode=ProcessMode.Repository,
            revision=self.worker_version.revision,
            creator=self.user,
            farm=Farm.objects.first(),
        )
        process.run()
        task = process.tasks.get()
        self.assertTrue(self.vol.type.folder)
        response = self.client.get(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            HTTP_AUTHORIZATION=f"Ponos {task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()["thumbnail_put_url"])

    def test_get_element_creator(self):
        self.vol.creator = self.user
        self.vol.save()
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["creator"], "Test user")

    def test_get_element_rights(self):
        cases = [
            (Role.Guest, self.superuser, ["read"]),
            (Role.Guest, self.user, ["read"]),
            (Role.Contributor, self.superuser, ["read", "write"]),
            (Role.Contributor, self.user, ["read", "write", "admin"]),
            (Role.Admin, self.superuser, ["read", "write", "admin"]),
            (Role.Admin, self.user, ["read", "write", "admin"]),
        ]
        self.client.force_login(self.user)
        for role, creator, expected_rights in cases:
            with self.subTest(role=role, creator=creator):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)
                self.vol.creator = creator
                self.vol.save()
                with self.assertNumQueries(6):
                    response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(response.json()["rights"], expected_rights)

    def test_get_element_created_by_worker_run(self):
        self.vol.worker_version = self.worker_version
        self.vol.worker_run = self.worker_run
        self.vol.save()

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.vol.id),
            "name": "Volume 1",
            "type": "volume",
            "corpus": {
                "id": str(self.corpus.id),
                "name": "Unit Tests",
                "public": True,
            },
            "thumbnail_url": self.vol.thumbnail.s3_url,
            "thumbnail_put_url": None,
            "worker_version": str(self.worker_version.id),
            "confidence": None,
            "zone": None,
            "rotation_angle": 0,
            "mirrored": False,
            "created": "2020-02-02T01:23:45.678000Z",
            "creator": None,
            "rights": ["read"],
            "metadata_count": 0,
            "classifications": [],
            "worker_run": {
                "id": str(self.worker_run.id),
                "summary": self.worker_run.summary
            }
        })

    def test_get_element_classification_worker_run(self):
        ml_class = MLClass.objects.create(name="text", corpus=self.corpus)
        classification = self.vol.classifications.create(worker_version=self.worker_version, worker_run=self.worker_run, ml_class=ml_class)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.vol.id),
            "name": "Volume 1",
            "type": "volume",
            "corpus": {
                "id": str(self.corpus.id),
                "name": "Unit Tests",
                "public": True,
            },
            "thumbnail_url": self.vol.thumbnail.s3_url,
            "thumbnail_put_url": None,
            "worker_version": None,
            "confidence": None,
            "zone": None,
            "rotation_angle": 0,
            "mirrored": False,
            "created": "2020-02-02T01:23:45.678000Z",
            "creator": None,
            "rights": ["read"],
            "metadata_count": 0,
            "classifications": [
                {
                    "id": str(classification.id),
                    "confidence": None,
                    "high_confidence": False,
                    "state": "pending",
                    "worker_version": str(self.worker_version.id),
                    "worker_run": {
                        "id": str(self.worker_run.id),
                        "summary": self.worker_run.summary
                    },
                    "ml_class": {
                        "id": str(ml_class.id),
                        "name": "text",
                    }
                }
            ],
            "worker_run": None
        })
