
import uuid

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestElementsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.act_type = cls.corpus.types.get(slug="act")
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.page = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.private_corpus = Corpus.objects.create(name="private", public=False)
        cls.private_page = Element.objects.create(
            type=ElementType.objects.create(corpus=cls.private_corpus, slug="page", display_name="Page"),
            corpus=cls.private_corpus,
            name="A Page (but private)",
        )

    def test_select_element_requires_login(self):
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_select_element_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_select_element_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["Selection is not available on this instance."])

    def test_select_element_wrong_corpus(self):
        user = User.objects.create_user("nope@nope.fr")
        user.verified_email = True
        user.save()
        self.client.force_login(user)
        element = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name="Act xx"
        )
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [element.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "ids": ["Some element IDs do not exist or you do not have permission to access them."],
        })

    def test_select_element_wrong_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id, "ffffffff-ffff-ffff-ffff-ffffffffffff"]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "ids": ["Some element IDs do not exist or you do not have permission to access them."],
        })

    def test_select_element_bad_uuid(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id, "oh no"]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertContains(response, "Must be a valid UUID", count=1, status_code=400)

    def test_select_duplicate_uuid(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id, self.page.id, self.vol.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        selected = self.client.get(reverse("api:elements-selection")).json()["results"]
        self.assertEqual(len(selected), 2)
        self.assertCountEqual([item["id"] for item in selected], [str(self.vol.id), str(self.page.id)])

    def test_select_element_no_id(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_select_element_not_a_list(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:elements-selection"), data={"ids": self.page.id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Expected a list of items but got type "str".'])

    def test_select_element(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        results = response.json()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.vol.id))
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.vol.id))

        response = self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id, self.page.id]},
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        results = response.json()
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result["id"] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result["id"] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )

    def test_delete_requires_login(self):
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        response = self.client.delete(reverse("api:elements-selection"), data={"id": self.page.id})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        response = self.client.delete(reverse("api:elements-selection"), data={"id": self.page.id})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_wrong_id(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        response = self.client.delete(reverse("api:elements-selection"), data={"id": self.page.id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_delete_not_a_uuid(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:elements-selection"), data={"id": {"id": self.page.id, "type": "page"}}, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"id": ["Not a valid UUID."]})

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_delete_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:elements-selection"), data={"id": self.page.id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["Selection is not available on this instance."])

    def test_delete_selection(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id, self.page.id]},
            format="json"
        )
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertCountEqual(
            [result["id"] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )

        with self.assertNumQueries(5):
            response = self.client.delete(reverse("api:elements-selection"), data={"id": self.vol.id})
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.page.id))

    def test_delete_selection_filter_corpus(self):
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.user.selected_elements.add(self.page, self.vol, self.private_page)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(reverse("api:elements-selection"), data={"corpus": self.private_corpus.id})
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]["id"], str(self.page.id))
        self.assertEqual(results[1]["id"], str(self.vol.id))

    def test_delete_selection_filter_corpus_bad_id(self):
        bad_id = uuid.uuid4()
        self.user.selected_elements.add(self.page, self.vol)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(reverse("api:elements-selection"), data={"corpus": bad_id})
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "Not found."})

    def test_delete_selection_filter_corpus_not_an_uuid(self):
        bad_id = "melon"
        self.user.selected_elements.add(self.page, self.vol)
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:elements-selection"), data={"corpus": bad_id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"corpus": "Not a valid UUID."})

    def test_select_element_two_session(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id]},
            format="json"
        )
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.vol.id))

        user = User.objects.create_user("nope@nope.fr")
        user.verified_email = True
        user.save()
        self.client.force_login(user)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.page.id]},
            format="json"
        )
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.page.id))

    def test_delete_all_selection(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.vol.id, self.page.id]},
            format="json"
        )
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result["id"] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )

        response = self.client.delete(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 0)

    def test_list_selection_requires_login(self):
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_selection_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_list_selection_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("api:elements-selection"))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["Selection is not available on this instance."])

    def test_list_selection(self):
        self.user.selected_elements.add(self.page, self.vol)
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:elements-selection"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]["id"], str(self.page.id))
        self.assertEqual(results[1]["id"], str(self.vol.id))

    def test_list_selection_filter_corpus(self):
        self.user.selected_elements.add(self.page, self.vol, self.private_page)
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:elements-selection"), data={"corpus": self.private_corpus.id})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.private_page.id))

    def test_list_selection_filter_corpus_bad_id(self):
        self.user.selected_elements.add(self.page, self.vol)
        bad_id = uuid.uuid4()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:elements-selection"), data={"corpus": bad_id})
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "Not found."})

    def test_list_selection_filter_corpus_not_an_uuid(self):
        self.user.selected_elements.add(self.page, self.vol)
        bad_id = 12
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:elements-selection"), data={"corpus": bad_id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"corpus": "Not a valid UUID."})
